@@ -105,7 +105,7 @@ struct blkdev {
 
     /* worker thread */
     pthread_t           mainthread;
-    
+
     /* request lists */
     struct list_head    inflight;
     struct list_head    finished;
@@ -121,7 +121,7 @@ struct blkdev {
     int                 cnt_aio;
     io_context_t        ctx;
 #endif
-    
+
     /* statistics */
     struct blkstats     st;
     struct blkstats     st_prev;
@@ -226,7 +226,7 @@ static int ioreq_parse(struct ioreq *ioreq)
 		 __FUNCTION__, ioreq->req.operation);
 	goto err;
     };
-    
+
     ioreq->start = ioreq->end = ioreq->req.sector_number * blkdev->file_blk;
     for (i = 0; i < ioreq->req.nr_segments; i++) {
 	if (i == BLKIF_MAX_SEGMENTS_PER_REQUEST) {
@@ -301,7 +301,7 @@ static int ioreq_map(struct ioreq *ioreq)
 {
     int gnt = ioreq->blkdev->xendev.gnttabdev;
     int i;
-    
+
     if (batch_maps) {
 	ioreq->pages = xc_gnttab_map_grant_refs
 	    (gnt, ioreq->vecs, ioreq->domids, ioreq->refs, ioreq->prot);
@@ -342,12 +342,12 @@ static int ioreq_runio_serial(struct ioreq *ioreq)
 {
     struct blkdev *blkdev = ioreq->blkdev;
     int len;
-    
+
     if (-1 == ioreq_map(ioreq))
 	goto err;
     if (ioreq->presync)
 	fdatasync(blkdev->file);
-	
+
     lseek(blkdev->file, ioreq->start, SEEK_SET);
     switch (ioreq->req.operation) {
     case BLKIF_OP_READ:
@@ -503,7 +503,7 @@ getmore:
 	    d1printf("%s: aio failure: op=%d\n", __FUNCTION__,
 		     events[i].obj->aio_lio_opcode);
 	}
-	
+
 	if (ioreq->iocbs_done < ioreq->iocbs)
 	    continue;
 
@@ -530,7 +530,7 @@ static int ioreq_runio_async_submit(struct ioreq *ioreq)
     struct iocb *iocb_list[IOCB_COUNT];
     struct io_event event;
     int i, rc, done = 0, try;
-    
+
     if (-1 == ioreq_map(ioreq))
 	goto err;
 
@@ -678,7 +678,7 @@ static int blk_send_response_one(struct ioreq *ioreq)
     resp.id        = ioreq->req.id;
     resp.operation = ioreq->req.operation;
     resp.status    = ioreq->status;
-    
+
     /* Place on the response ring for the relevant domain. */
     switch (blkdev->protocol) {
     case BLKIF_PROTOCOL_NATIVE:
@@ -731,7 +731,7 @@ static void blk_send_response_all(struct blkdev *blkdev)
     if (send_notify)
 	xc_evtchn_notify(blkdev->xendev.evtchnd, blkdev->xendev.local_port);
 }
-	
+
 static int blk_get_request(struct blkdev *blkdev, struct ioreq *ioreq, RING_IDX rc)
 {
     switch (blkdev->protocol) {
@@ -783,7 +783,7 @@ static void *blk_thread_main(void *arg)
 	    break;
 	rc = blkdev->rings.common.req_cons;
 	rp = blkdev->rings.common.sring->req_prod;
-	rmb(); /* Ensure we see queued requests up to 'rp'. */
+	xen_rmb(); /* Ensure we see queued requests up to 'rp'. */
 
 	/* Limit #of requests we queue up for I/O so we ack requests
 	 * faster if busy.  Improves backend/frontend parallelism and
@@ -792,7 +792,7 @@ static void *blk_thread_main(void *arg)
 	    rp = rc + (max_requests >> 2);
 	    blkdev->more_work++;
 	}
-	
+
 	while ((rc != rp)) {
 	    /* pull request from ring */
 	    if (RING_REQUEST_CONS_OVERFLOW(&blkdev->rings.common, rc))
@@ -843,7 +843,7 @@ static void *blk_thread_main(void *arg)
 	ioreq_runio_async_getevents(blkdev, 1);
 #endif
     blk_send_response_all(blkdev);
-    
+
     d1printf("%s: exit\n", __FUNCTION__);
     return NULL;
 }
@@ -899,12 +899,12 @@ static void blk_setup_backend(struct blkdev *blkdev)
 {
     int mode, qflags, use_qemu, have_barriers, info = 0;
     char *h;
-    
+
     if (blkdev->xendev.state != XENDEV_PROBED)
 	return;
     if (!blkdev->xendev.online)
 	return;
-	
+
     /* read xenstore entries */
     if (NULL == blkdev->params) {
 	blkdev->params = read_be_str(&blkdev->xendev, "params");
@@ -993,7 +993,7 @@ static void blk_setup_backend(struct blkdev *blkdev)
 	     use_qemu        ? "yes" : "no",
 	     blkdev->use_aio ? "yes" : "no",
 	     blkdev->file_size, blkdev->file_size >> 20);
-    
+
     /* fill info */
     write_be_int(&blkdev->xendev, "feature-barrier", have_barriers);
     write_be_int(&blkdev->xendev, "info",            info);
@@ -1020,10 +1020,12 @@ static void blk_connect_frontend(struct blkdev *blkdev)
 	return;
 
     blkdev->protocol = BLKIF_PROTOCOL_NATIVE;
-    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_32))
-	blkdev->protocol = BLKIF_PROTOCOL_X86_32;
-    if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_64))
-	blkdev->protocol = BLKIF_PROTOCOL_X86_64;
+    if (blkdev->xendev.protocol) {
+        if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_32))
+            blkdev->protocol = BLKIF_PROTOCOL_X86_32;
+        if (0 == strcmp(blkdev->xendev.protocol, XEN_IO_PROTO_ABI_X86_64))
+            blkdev->protocol = BLKIF_PROTOCOL_X86_64;
+    }
 
     blkdev->sring = xc_gnttab_map_grant_ref(blkdev->xendev.gnttabdev,
 					    blkdev->xendev.dom,
@@ -1076,7 +1078,7 @@ static void blk_disconnect(struct blkdev *blkdev, enum xenbus_state state)
     void *dummy;
 
     if (blkdev->xendev.state == XENDEV_DISCONNECTED)
-	return;
+        goto out;
     blkdev->xendev.state = XENDEV_DISCONNECTED;
 
     if (blkdev->mainthread) {
@@ -1087,6 +1089,7 @@ static void blk_disconnect(struct blkdev *blkdev, enum xenbus_state state)
 #ifdef HAVE_LIBAIO
     if (blkdev->use_aio) {
 	io_destroy(blkdev->ctx);
+        memset(&blkdev->ctx, 0, sizeof(blkdev->ctx));
 	blkdev->use_aio = 0;
     }
 #endif
@@ -1111,6 +1114,7 @@ static void blk_disconnect(struct blkdev *blkdev, enum xenbus_state state)
 	blkdev->sring = NULL;
     }
 
+out:
     if (state)
 	change_state_xendev(&blkdev->xendev, state);
 }
@@ -1164,6 +1168,14 @@ static int blk_frontend(struct xendev *xendev, char *node, char *val)
     d2printf("%s: %p \"%s\" = \"%s\"\n", __FUNCTION__, blkdev, node,
 	     val ? val : "<deleted>");
     switch (xendev->fe_state) {
+    case XenbusStateInitialising:
+        if (blkdev->xendev.state == XENDEV_DISCONNECTED) {
+            /* re-init backend for other frontend */
+            blkdev->xendev.state = XENDEV_PROBED;
+            blk_setup_backend(blkdev);
+        }
+        break;
+
     case XenbusStateInitialised:
     case XenbusStateConnected:
 	blk_connect_frontend(blkdev);
@@ -1175,7 +1187,6 @@ static int blk_frontend(struct xendev *xendev, char *node, char *val)
 	blk_disconnect(blkdev, xendev->fe_state);
 	break;
 
-    case XenbusStateInitialising:
     case XenbusStateInitWait:
     case XenbusStateUnknown:
 	/* make gcc happy */
@@ -1277,7 +1288,7 @@ static int blk_free(struct xendev *xendev)
 	list_del(&ioreq->list);
 	free(ioreq);
     }
-	
+
     if (blkdev->params)
 	free(blkdev->params);
     if (blkdev->mode)
