@@ -9,19 +9,25 @@ sreg_t error_noperm(struct xen_cpu *cpu, ureg_t *args)
 
 sreg_t console_io(struct xen_cpu *cpu, ureg_t *args)
 {
-    if (args[0] == CONSOLEIO_write) {
-	void *ptr = (void*)args[1];
-	int count = args[2];
-	uint8_t buf[128];
+    int count = args[1];
+    void *ptr = (void*)args[2];
+    uint8_t buf[128];
 
-	if (count > sizeof(buf)-1)
+    switch (args[0]) {
+    case CONSOLEIO_write:
+ 	if (count > sizeof(buf)-1)
 	    count = sizeof(buf)-1;
 	if (0 != memcpy_pf(&buf, ptr, count))
 	    return -EFAULT;
 	buf[count] = 0;
+        while (count > 0 && (buf[count-1] == '\r' || buf[count-1] == '\n'))
+            buf[--count] = 0;
 	printk(1, "guest: \"%s\"\n", buf);
 	return count;
-    } else {
+    case CONSOLEIO_read:
+        return 0;
+    default:
+        printk(1, "console: unknown: %s\n", consoleio_name(args[0]));
 	return -ENOSYS;
     }
 }
@@ -72,7 +78,7 @@ sreg_t update_descriptor(struct xen_cpu *cpu, ureg_t *args)
 	index = addr_offset(pa) / sizeof(struct descriptor_32);
 	cpu->gdt[p * 512 + index] = desc;
     }
-    
+
     /* update guest gdt/ldt */
     guest_gdt = map_page(pa);
     *guest_gdt = desc;
@@ -96,7 +102,7 @@ sreg_t grant_table_op(struct xen_cpu *cpu, ureg_t *args)
     struct gnttab_query_size  *qs;
     unsigned long *frames;
     int i, rc = 0;
-    
+
     switch (args[0]) {
     case GNTTABOP_setup_table:
 	st = (void*)args[1];
@@ -128,7 +134,7 @@ sreg_t grant_table_op(struct xen_cpu *cpu, ureg_t *args)
 sreg_t xen_version(struct xen_cpu *cpu, ureg_t *args)
 {
     static const char extra[XEN_EXTRAVERSION_LEN] = "-xenner-" VERSION;
-    
+
     switch (args[0]) {
     case XENVER_version:
 	return (3 << 16) | 1;
@@ -136,6 +142,25 @@ sreg_t xen_version(struct xen_cpu *cpu, ureg_t *args)
 	if (0 != memcpy_pf((void*)args[1], extra, sizeof(extra)))
 	    return -EFAULT;
 	return 0;
+    case XENVER_capabilities:
+    {
+#if longmode
+        char caps[] = "xen-3.0-x86_64";
+	if (0 != memcpy_pf((void*)args[1], caps, sizeof(caps)))
+	    return -EFAULT;
+#else
+        char caps_pae[] = "xen-3.0-x86_32p";
+        char caps_32[]  = "xen-3.0-x86_32";
+        if (is_pae()) {
+            if (0 != memcpy_pf((void*)args[1], caps_pae, sizeof(caps_pae)))
+                return -EFAULT;
+        } else {
+            if (0 != memcpy_pf((void*)args[1], caps_32, sizeof(caps_32)))
+                return -EFAULT;
+        }
+#endif
+        break;
+    }
 
     case XENVER_get_features:
     {
@@ -170,10 +195,14 @@ sreg_t vm_assist(struct xen_cpu *cpu, ureg_t *args)
     case VMASST_CMD_enable:
 	printk(1, "%s: enable %d (%s)\n", __FUNCTION__,
 	       type, vmasst_type_name(type));
+        if (type == VMASST_TYPE_writable_pagetables)
+            wrpt = 1;
 	break;
     case VMASST_CMD_disable:
 	printk(1, "%s: disable %d (%s)\n", __FUNCTION__,
 	       type, vmasst_type_name(type));
+        if (type == VMASST_TYPE_writable_pagetables)
+            wrpt = 0;
 	break;
     default:
 	printk(0, "%s: FIXME: unknown %d\n", __FUNCTION__, (int)args[0]);
@@ -193,7 +222,7 @@ sreg_t sched_op(struct xen_cpu *cpu, ureg_t *args)
     case SCHEDOP_block:
 	guest_sti(cpu);
 	if (!evtchn_pending(cpu)) {
-	    sti(); halt(cpu->id); cli();
+            halt_i(cpu->id);
 	    pv_clock_update(1);
 	}
 	break;
@@ -232,7 +261,7 @@ sreg_t sched_op_compat(struct xen_cpu *cpu, ureg_t *args)
 sreg_t memory_op(struct xen_cpu *cpu, ureg_t *args)
 {
     int cmd = args[0] & 0x0f /* MEMOP_CMD_MASK */;
-    
+
     switch (cmd) {
     case XENMEM_increase_reservation:
     {
@@ -285,18 +314,16 @@ sreg_t memory_op(struct xen_cpu *cpu, ureg_t *args)
 	if (res.domid != DOMID_SELF)
 	    return -EPERM;
 	ptr = res.extent_start.p;
+        gmfn = vmconf.mfn_guest;
 	for (i = 0; i < res.nr_extents; i++) {
 	    if (0 != memcpy_pf(&gpfn, ptr + i, sizeof(gpfn)))
 		break;
-	    gmfn = gpfn + vmconf.mfn_guest;
 	    for (p = 0; p < (1 << res.extent_order); p++) {
-		if (m2p[gmfn+p] != INVALID_M2P_ENTRY) {
-		    /* trying to balloon up might trigger this */
-		    printk(1, "%s:    %4d: gpfn %08lx, gmfn %08lx: %s\n", __FUNCTION__,
-			   i, gpfn+p, gmfn+p, "Huh? Page isn't free!");
-		    break;
-		}
-		m2p[gmfn+p] = gpfn+p;
+                while (m2p[gmfn] != INVALID_M2P_ENTRY && gmfn < vmconf.pg_total)
+                    gmfn++;
+                if (gmfn == vmconf.pg_total)
+                    break;
+		m2p[gmfn] = gpfn+p;
 	    }
 	    if (p != (1 << res.extent_order))
 		break;
@@ -330,7 +357,7 @@ sreg_t memory_op(struct xen_cpu *cpu, ureg_t *args)
 	    return -EFAULT;
 	return 0;
     }
-    
+
     case XENMEM_memory_map:
 	/* we have no e820 map */
 	return -ENOSYS;
@@ -349,7 +376,7 @@ sreg_t set_trap_table(struct xen_cpu *cpu, ureg_t *args)
     int i;
 
     if (0 == args[0]) {
-	printk(0, "%s: FIXME: clear table\n", __FUNCTION__);
+        memset(&xentr, 0, sizeof(xentr));
 	return -EINVAL;
     }
 
@@ -386,7 +413,7 @@ static int callback_setup(int type, xen_callback_t *cb)
 	ok = 1;
 	break;
     }
-    
+
     printk(1, "%s: %s (#%d) -> %s\n", __FUNCTION__,
 	   callbacktype_name(type), type,
 	   ok ? "OK" : "unsupported");
@@ -436,7 +463,7 @@ sreg_t callback_op(struct xen_cpu *cpu, ureg_t *args)
     memcpy_pf(&cb, (void*)(args[1]), sizeof(cb));
     if (cb.type >= 8)
 	return -EINVAL;
-    
+
     switch (args[0]) {
     case CALLBACKOP_register:
 	if (0 != callback_setup(cb.type, &cb.address))
@@ -550,7 +577,7 @@ sreg_t vcpu_op(struct xen_cpu *cpu, ureg_t *args)
     case VCPUOP_initialise:
     {
 	struct xen_cpu *vcpu;
-	
+
 	printk(0, "%s: initialise cpu %d\n", __FUNCTION__, (int)args[1]);
 	vcpu = cpu_find(args[1]);
 	if (!vcpu->init_ctxt)
@@ -563,7 +590,7 @@ sreg_t vcpu_op(struct xen_cpu *cpu, ureg_t *args)
     case VCPUOP_up:
     {
 	struct xen_cpu *vcpu;
-	
+
 	printk(0, "%s: up cpu %d\n", __FUNCTION__, (int)args[1]);
 	vcpu = cpu_find(args[1]);
 	lapic_ipi_boot(cpu, vcpu);
@@ -590,7 +617,7 @@ sreg_t vcpu_op(struct xen_cpu *cpu, ureg_t *args)
 	if (!pv_have_clock)
 	    return -ENOSYS;
 
-	vcpu->v.vcpu_page = fixmap_page(frame_to_addr(reg.mfn));
+	vcpu->v.vcpu_page = fixmap_page(cpu, frame_to_addr(reg.mfn));
 	new_info = vcpu->v.vcpu_page + reg.offset;
 	new_info_pa = frame_to_addr(reg.mfn) + reg.offset;
 	printk(1,"%s/%d: vcpu_info: mfn 0x%" PRIx64 ", offset 0x%x, pa %" PRIx64 " mapped to %p\n",
@@ -615,7 +642,7 @@ sreg_t set_timer_op(struct xen_cpu *cpu, ureg_t *args)
 #else
     uint64_t time = args[0] | (uint64_t)args[1] << 32;
 #endif
-    
+
     cpu->oneshot = time;
     lapic_timer(cpu);
     return 0;
@@ -717,6 +744,15 @@ sreg_t event_channel_op(struct xen_cpu *cpu, ureg_t *args)
 	evtchn_unmask(cpu, unmask.port);
 	return 0;
     }
+    case EVTCHNOP_close:
+    {
+	struct evtchn_close cl;
+
+	if (0 != memcpy_pf(&cl, (void*)args[1], sizeof(cl)))
+	    return -EFAULT;
+	evtchn_close(cpu, cl.port);
+	return 0;
+    }
     default:
 	return -ENOSYS;
     }
@@ -757,7 +793,6 @@ sreg_t mmuext_op(struct xen_cpu *cpu, ureg_t *args)
 	    /* ignore */
 	    break;
 	case MMUEXT_UNPIN_TABLE:
-	    vminfo.faults[XEN_FAULT_TMP_8]++;
 	    /* KVM_MMU_OP_RELEASE_PT ??? */
 	    break;
 	case MMUEXT_INVLPG_LOCAL:
@@ -806,19 +841,17 @@ sreg_t mmuext_op(struct xen_cpu *cpu, ureg_t *args)
 
 #if longmode
 	case MMUEXT_NEW_BASEPTR:
-	    cpu->kernel_cr3_mfn = uops->arg1.mfn;
-	    update_emu_mappings(cpu->kernel_cr3_mfn);
-	    pv_write_cr3(cpu, frame_to_addr(cpu->kernel_cr3_mfn));
-	    vminfo.faults[XEN_FAULT_OTHER_CR3_LOAD]++;
+	    update_emu_mappings(uops->arg1.mfn);
+	    pv_write_cr3(cpu, uops->arg1.mfn);
+            break;
 	case MMUEXT_NEW_USER_BASEPTR:
+	    update_emu_mappings(uops->arg1.mfn);
 	    cpu->user_cr3_mfn = uops->arg1.mfn;
-	    update_emu_mappings(cpu->user_cr3_mfn);
 	    break;
 #else
 	case MMUEXT_NEW_BASEPTR:
 	    update_emu_mappings(uops->arg1.mfn);
-	    write_cr3(frame_to_addr(uops->arg1.mfn));
-	    vminfo.faults[XEN_FAULT_OTHER_CR3_LOAD]++;
+	    pv_write_cr3(cpu, uops->arg1.mfn);
 	    break;
 #endif
 	default:
