@@ -1,23 +1,39 @@
 
 void *memset(void *s, int c, size_t n)
 {
-	uint8_t *p = s;
-	uint32_t i;
+    uint8_t *p = s;
+    uint32_t i;
 
-	for (i = 0; i < n; i++)
-		p[i] = c;
-	return s;
+    for (i = 0; i < n; i++)
+        p[i] = c;
+    return s;
 }
 
 void *memcpy(void *dest, const void *src, size_t n)
 {
-	const uint8_t *s = src;
-	uint8_t *d = dest;
-	uint32_t i;
+    const uint8_t *s = src;
+    uint8_t *d = dest;
+    uint32_t i;
 
-	for (i = 0; i < n; i++)
-		d[i] = s[i];
-	return dest;
+    for (i = 0; i < n; i++)
+        d[i] = s[i];
+    return dest;
+}
+
+int memcmp(const void *s1, const void *s2, size_t n)
+{
+    const uint8_t *a = s1;
+    const uint8_t *b = s2;
+    int i;
+
+    for (i = 0; i < n; i++) {
+        if (a[i] == b[i])
+            continue;
+        if (a[i] < b[i])
+            return -1;
+        return 1;
+    }
+    return 0;
 }
 
 /* --------------------------------------------------------------------- */
@@ -75,7 +91,7 @@ int panic(char *message, struct regs *regs)
     printk(0, "panic: %s\n", message);
     if (regs)
 	print_state(regs);
-    vmexit_panic();
+    emudev_cmd(EMUDEV_CMD_GUEST_SHUTDOWN, -1);
     return 0;
 }
 
@@ -140,7 +156,7 @@ void print_registers(int level, struct regs *regs)
 void print_stack(int level, ureg_t rsp)
 {
     ureg_t max;
-    
+
     max = ((rsp + PAGE_SIZE) & PAGE_MASK) - sizeof(ureg_t);
     printk(level, "printing stack " DR " - " DR "\n", rsp, max);
     while (rsp <= max) {
@@ -167,7 +183,7 @@ static struct descriptor_32 mkdesc(const struct kvm_segment *seg)
     shift  = seg->g ? 12 : 0;
     desc.a = (seg->limit >> shift) & 0xffff;
     desc.b = (seg->limit >> shift) & 0x000f0000;
-    
+
     desc.a |= (seg->base & 0xffff) << 16;
     desc.b |= seg->base & 0xff000000;
     desc.b |= (seg->base & 0xff0000) >> 16;
@@ -307,7 +323,7 @@ void guest_cpu_init(struct xen_cpu *cpu)
     ctxt->user_regs.fs = fix_sel(ctxt->user_regs.fs);
     ctxt->user_regs.gs = fix_sel(ctxt->user_regs.gs);
     ctxt->user_regs.ss = fix_sel(ctxt->user_regs.ss);
-    
+
     cpu->kernel_ss = ctxt->kernel_ss;
     cpu->kernel_sp = ctxt->kernel_sp;
 }
@@ -325,9 +341,10 @@ static void guest_start_info(struct xen_cpu *cpu, struct regs *regs)
 static struct xen_cpu *cpu_alloc(int id)
 {
     struct xen_cpu *cpu;
-    
+    ureg_t cr3;
+
     printk(1, "%s: cpu %d\n", __FUNCTION__, id);
-    
+
     cpu = get_memory(sizeof(*cpu), "per-cpu data");
     cpu->id = id;
     cpu->periodic = XEN_DEFAULT_PERIOD;
@@ -335,7 +352,10 @@ static struct xen_cpu *cpu_alloc(int id)
     cpu->v.vcpu_info_pa = EMU_PA(cpu->v.vcpu_info);
     guest_cli(cpu);
     list_add_tail(&cpu->next, &cpus);
-    
+
+    asm volatile("mov %%cr3,%0" : "=r" (cr3));
+    pv_write_cr3(cpu, addr_to_frame(cr3));
+
     gdt_init(cpu);
     stack_init(cpu);
     tss_init(cpu);
@@ -409,7 +429,7 @@ static void userspace_config(void)
 asmlinkage void do_boot(struct regs *regs)
 {
     struct xen_cpu *cpu;
-    
+
     printk(0, "this is %s (xenner %s), boot cpu #0\n", EMUNAME, VERSION);
 
     userspace_config();
@@ -421,7 +441,7 @@ asmlinkage void do_boot(struct regs *regs)
     cpu_init(cpu);
     printk(1, "%s: boot cpu setup done\n", EMUNAME);
 
-    paging_init();
+    paging_init(cpu);
     paging_start(cpu);
     printk(1, "%s: paging setup done\n", EMUNAME);
 
@@ -475,11 +495,11 @@ asmlinkage void do_illegal_instruction(struct regs *regs)
 	bounce_trap(cpu, regs, 6, -1);
 	return;
     }
-    
+
     skip = emulate(cpu, regs);
     switch (skip) {
     case -1: /* error */
-	panic("instruction emulation failed\n", regs);
+	panic("instruction emulation failed (ill)\n", regs);
 	break;
     case 0:  /* bounce to guest */
 	bounce_trap(cpu, regs, 6, -1);
@@ -518,7 +538,7 @@ static int is_allowed_io(struct xen_cpu *cpu, struct regs *regs)
     }
     return 0; /* no */
 }
-    
+
 asmlinkage void do_general_protection(struct regs *regs)
 {
     struct xen_cpu *cpu = get_cpu();
@@ -539,7 +559,7 @@ asmlinkage void do_general_protection(struct regs *regs)
 	bounce_trap(cpu, regs, 13, -1);
 	return;
     }
-    
+
     if (0 != regs->error) {
 	print_gpf_info(0, cpu, regs);
 	panic("unhandled kernel gpf", regs);
@@ -549,7 +569,8 @@ emulate:
     skip = emulate(cpu, regs);
     switch (skip) {
     case -1: /* error */
-	panic("instruction emulation failed", regs);
+	print_gpf_info(0, cpu, regs);
+	panic("instruction emulation failed (gpf)", regs);
 	break;
     case 0:  /* bounce to guest */
 	vminfo.faults[XEN_FAULT_GENERAL_PROTECTION_GUEST]++;
@@ -637,13 +658,13 @@ asmlinkage void do_smp_flush_tlb(struct regs *regs)
 void flush_tlb_remote(struct xen_cpu *cpu, ureg_t mask, ureg_t addr)
 {
     int cpus;
-    
+
     mask &= ~(1 << cpu->id);
     if (!mask) {
 	vminfo.faults[XEN_FAULT_OTHER_FLUSH_TLB_NONE]++;
 	return;
     }
-    
+
     /*
      * we must be able to process ipi while waiting for the lock,
      * otherwise we deadlock in case another cpu busy-waits for us
