@@ -3,7 +3,7 @@ void *get_pages(int pages, char *purpose)
 {
     static void *heap_top;
     void *ptr;
-    
+
     if (!heap_top)
 	heap_top = _vstop;
     ptr = heap_top;
@@ -26,5 +26,5 @@ void paging_start(struct xen_cpu *cpu)
 
     cr3_mfn = cpu->init_ctxt->ctrlreg[3] >> PAGE_SHIFT;
     update_emu_mappings(cr3_mfn);
-    write_cr3(frame_to_addr(cr3_mfn));
+    pv_write_cr3(cpu, cr3_mfn);
 }
