@@ -88,7 +88,7 @@ static void sigsetup(struct xenvm *xen)
 int kvmbits_signal_vcpus(struct xenvm *xen, int signal, int self)
 {
     int i;
-    
+
     for (i = 0; i < xen->vcpus; i++) {
 	if (!xen->vcpu[i].thread)
 	    continue;
@@ -104,7 +104,7 @@ void alive(void)
     static const char alive[] = ".oOo";
     static const int chars = sizeof(alive)/sizeof(alive[0]) -1;
     static int count;
-    
+
     fprintf(stderr, "[%c]\r", alive[count++ % chars]);
 }
 
@@ -174,8 +174,11 @@ void update_time(struct xenvcpu *vcpu)
     xentime->tsc_to_system_mul = xen->tsc_mul_frac;
     xentime->tsc_shift         = xen->tsc_shift;
     xentime->version++;
+
+#if 0
     d3printf("%d: systime %" PRId64 " tsc_timestamp %" PRId64 "\n", vcpu->id,
 	     xentime->system_time, xentime->tsc_timestamp);
+#endif
 
 #if 1 /* FIXME: move elsewhere */
     xen_wc->wc_version++;
@@ -228,7 +231,7 @@ static void check_events(struct xenvm *xen)
 	    once = 0;
 	}
     }
-    
+
     if (0 /* FIXME */)
 	balloon(xen);
 }
@@ -293,7 +296,7 @@ void xenner_cleanup(struct xenvm *xen, int stats)
 static void print_stack_32(struct xenvm *xen, char *name, uint32_t esp, uint32_t *ptr)
 {
     int i;
-    
+
     logprintf(xen, "--- %s stack trace ---\n", name);
     for (i = 0; i < 512; i++) {
 	if (0 == addr_offset(esp + 4*i))
@@ -311,7 +314,7 @@ static void print_stack_32(struct xenvm *xen, char *name, uint32_t esp, uint32_t
 static void print_stack_64(struct xenvm *xen, char *name, uint64_t esp, uint64_t *ptr)
 {
     int i;
-    
+
     logprintf(xen, "--- %s stack trace ---\n", name);
     for (i = 0; i < 256; i++) {
 	if (0 == addr_offset(esp + 8*i))
@@ -433,7 +436,7 @@ void vm_kill(struct xenvcpu *vcpu, const char *reason, uint64_t guest_rsp)
 	/* dump current vcpu only */
 	dump_vcpu(vcpu, guest_rsp);
     }
-    
+
     xenner_cleanup(xen, 1);
     exit(1);
 }
@@ -447,7 +450,14 @@ static void do_emudev_command(struct xenvm *xen, uint16_t cmd, uint16_t arg)
 	/* nop vmexit (for time update without pvclock) */
 	break;
     case EMUDEV_CMD_WRITE_CHAR:
-	fprintf(stderr, "%c", arg);
+        if (xen->elen < sizeof(xen->eline)-1)
+            xen->eline[xen->elen++] = arg;
+        if (arg == '\r' || arg == '\n') {
+            xen->eline[xen->elen] = 0;
+            if (xen->elen > 1)
+                conemu_print(xen, xen->eline);
+            xen->elen = 0;
+        }
 	break;
     case EMUDEV_CMD_CONFIGURATION_DONE:
 	/* emu finisted initial EMUDEV_CONF_* setup, process now */
@@ -465,6 +475,9 @@ static void do_emudev_command(struct xenvm *xen, uint16_t cmd, uint16_t arg)
     case EMUDEV_CMD_EVTCHN_SEND:
 	evtchn_notify(xen, arg);
 	break;
+    case EMUDEV_CMD_EVTCHN_CLOSE:
+	evtchn_close(xen, arg);
+	break;
     case EMUDEV_CMD_INIT_SECONDARY_VCPU:
     {
 	struct xenvcpu *vcpu;
@@ -626,7 +639,7 @@ static int cb_outl(void *opaque, uint16_t addr, uint32_t value)
 static int cb_debug(void *opaque, int vcpu_id)
 {
     struct xenvm *xen = opaque;
-    
+
     d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
     return 0;
 }
@@ -674,7 +687,6 @@ static void cb_post_kvm_run(void *opaque, int vcpu_id)
     struct xenvm *xen = opaque;
 
     pthread_mutex_lock(&xen->biglock);
-    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
 }
 
 static int cb_pre_kvm_run(void *opaque, int vcpu_id)
@@ -682,8 +694,6 @@ static int cb_pre_kvm_run(void *opaque, int vcpu_id)
     struct xenvm *xen = opaque;
     struct xenvcpu *vcpu = xen->vcpu + vcpu_id;
 
-    d3printf("callback: %s/%d\n", __FUNCTION__, vcpu_id);
-
     if (!(xen->features & (1 << KVM_FEATURE_CLOCKSOURCE)))
 	update_time(vcpu);
 
@@ -740,7 +750,7 @@ static struct kvm_para_features {
 int kvmbits_features_enable(char *name, int state)
 {
     int i;
-    
+
     for (i = 0; i < sizeof(para_features)/sizeof(para_features[0]); i++) {
 	if (0 != strcmp(name, para_features[i].name))
 	    continue;
@@ -796,88 +806,6 @@ static struct kvm_callbacks xenner_cb = {
     .pre_kvm_run         = cb_pre_kvm_run,
 };
 
-int kvmbits_init(struct xenvm *xen)
-{
-    int i;
-    
-    /* init kvm */
-    xen->kvm = kvm_init(&xenner_cb, xen);
-    if (!xen->kvm) {
-	logprintf(xen, "kvm_init failed\n");
-	goto err;
-    }
-
-    if (kvmbits_features_check(xen) < 0)
-	goto err;
-    
-    /* create vm */
-    xen->pg_total &= ~((1<<10)-1);  /* round to 4MB border */
-    if (vmcore_init(xen) < 0)
-	goto err;
-    if (kvm_create_vm(xen->kvm) < 0) {
-	logprintf(xen, "kvm_create_vm failed\n");
-	goto err;
-    }
-
-    if (kvm_set_tss_addr(xen->kvm, xen->pg_total * PAGE_SIZE) < 0) {
-	logprintf(xen, "kvm_set_tss_addr failed (%s)\n", strerror(errno));
-	goto err;
-    }
-
-    if (kvm_register_userspace_phys_mem(xen->kvm, 0, xen->memory,
-					xen->pg_total * PAGE_SIZE, 1) < 0) {
-	logprintf(xen, "kvm_register_userspace_phys_mem failed\n");
-	goto err;
-    }
-
-    kvm_create_irqchip(xen->kvm);
-    if (!kvm_irqchip_in_kernel(xen->kvm)) {
-	logprintf(xen, "kvm_create_irqchip failed\n");
-	goto err;
-    }
-
-    for (i = 0; i < xen->vcpus; i++) {
-	if (kvm_create_vcpu(xen->kvm, i) < 0) {
-	    logprintf(xen, "kvm_create_vcpu failed\n");
-	    goto err;
-	}
-	xen->vcpu[i].id = i;
-	xen->vcpu[i].vm = xen;
-	need_regs(xen->vcpu + i);
-	need_sregs(xen->vcpu + i);
-    }
-
-    banner_print(xen, "setup");
-    sigsetup(xen);
-    if (0 != evtchn_init(xen))
-	goto err;
-    if (0 != xenstore_init_early(xen))
-	goto err;
-    console_init(xen);
-
-    if (0 != domain_builder(xen)) {
-	logprintf(xen, "domain builder failed\n");
-	goto err;
-    }
-
-    xenstore_init_late(xen);
-    setup_cpuid(xen);
-
-    tsc_calibrate(xen, &xen->tsc_shift, &xen->tsc_mul_frac);
-    xen->boot = get_systime();
-
-    return 0;
-
-err:
-    if (xen->kvm)
-	kvm_finalize(xen->kvm);
-    if (-1 != xen->corefd)
-	unlink(xen->corefile);
-    return -1;
-}
-
-/* ------------------------------------------------------------------ */
-
 static void *vcpu_thread(void *arg)
 {
     struct xenvcpu *vcpu = arg;
@@ -885,8 +813,7 @@ static void *vcpu_thread(void *arg)
 
     sigprocmask(SIG_BLOCK, &main_signals, NULL);
     vcpu->thread_id = syscall(SYS_gettid);
-    if (xen->start_stopped)
-	pthread_kill(pthread_self(), SIGUSR1);
+    pthread_kill(pthread_self(), SIGUSR1);
 
     pthread_mutex_lock(&xen->biglock);
     d1printf("%s/%d: start\n", __FUNCTION__, vcpu->id);
@@ -898,6 +825,149 @@ static void *vcpu_thread(void *arg)
     return NULL;
 }
 
+int kvmbits_init1(struct xenvm *xen)
+{
+    struct xenvcpu *vcpu;
+    int i;
+
+    /* init kvm */
+    xen->kvm = kvm_init(&xenner_cb, xen);
+    if (!xen->kvm) {
+	logprintf(xen, "kvm_init failed\n");
+	goto err;
+    }
+
+    if (kvmbits_features_check(xen) < 0)
+	goto err;
+
+    /* create vm1 */
+    xen->pg_total &= ~((1<<10)-1);  /* round to 4MB border */
+    if (vmcore_init1(xen) < 0)
+	goto err;
+
+    if (kvm_create_vm(xen->kvm) < 0) {
+	logprintf(xen, "kvm_create_vm failed\n");
+	goto err;
+    }
+
+    if (kvm_set_tss_addr(xen->kvm, xen->pg_total * PAGE_SIZE) < 0) {
+	logprintf(xen, "kvm_set_tss_addr failed (%s)\n", strerror(errno));
+	goto err;
+    }
+
+    kvm_create_irqchip(xen->kvm);
+    if (!kvm_irqchip_in_kernel(xen->kvm)) {
+	logprintf(xen, "kvm_create_irqchip failed\n");
+	goto err;
+    }
+
+    sigsetup(xen);
+    for (i = 0; i < xen->vcpus; i++) {
+	if (kvm_create_vcpu(xen->kvm, i) < 0) {
+	    logprintf(xen, "kvm_create_vcpu failed\n");
+	    goto err;
+	}
+	vcpu = xen->vcpu + i;
+        vcpu->id = i;
+	vcpu->vm = xen;
+	need_regs(vcpu);
+	need_sregs(vcpu);
+	pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);
+    }
+
+    if (0 != xenstore_init_early(xen))
+	goto err;
+    return 0;
+
+err:
+    if (xen->kvm)
+	kvm_finalize(xen->kvm);
+    if (-1 != xen->corefd)
+	unlink(xen->corefile);
+    return -1;
+}
+
+static int kvmbits_poll_early(struct xenvm *xen)
+{
+    struct list_head *item;
+    struct xenner_callback *cb;
+    struct timeval tv;
+    fd_set rd;
+    int count, rc, max;
+
+    for (count = 0; count < 16; count++) {
+	FD_ZERO(&rd);
+	max = 0;
+	list_for_each(item, &xen->callbacks) {
+	    cb = list_entry(item, struct xenner_callback, list);
+	    FD_SET(cb->fd,&rd);
+	    if (max < cb->fd)
+		max = cb->fd;
+	}
+
+        tv.tv_sec = 0;
+        tv.tv_usec = 500 * 1000; /* 0.5 sec */
+	rc = select(max+1, &rd, NULL, NULL, &tv);
+        if (rc < 0)
+            continue;
+        if (rc == 0)
+            break;
+
+	list_for_each(item, &xen->callbacks) {
+	    cb = list_entry(item, struct xenner_callback, list);
+            if (FD_ISSET(cb->fd,&rd))
+		cb->callback(xen, cb->fd, cb->cb_data);
+	}
+    }
+    return 0;
+}
+
+/*
+ * the stuff which can take some time ...
+ */
+int kvmbits_init2(struct xenvm *xen, int populate)
+{
+    kvmbits_poll_early(xen);
+
+    if (vmcore_init2(xen, populate) < 0)
+	goto err;
+    kvmbits_poll_early(xen);
+
+    if (kvm_register_userspace_phys_mem(xen->kvm, 0, xen->memory,
+					xen->pg_total * PAGE_SIZE, 1) < 0) {
+	logprintf(xen, "kvm_register_userspace_phys_mem failed\n");
+	goto err;
+    }
+
+    banner_print(xen, "setup");
+    if (0 != evtchn_init(xen))
+	goto err;
+    xen->xenstore_event = evtchn_port(xen, "xenstore");
+    console_init(xen);
+    kvmbits_poll_early(xen);
+
+    if (0 != domain_builder(xen)) {
+	logprintf(xen, "domain builder failed\n");
+	goto err;
+    }
+    kvmbits_poll_early(xen);
+
+    xenstore_init_late(xen);
+    setup_cpuid(xen);
+
+    tsc_calibrate(xen, &xen->tsc_shift, &xen->tsc_mul_frac);
+    xen->boot = get_systime();
+
+    return 0;
+
+err:
+    kvm_finalize(xen->kvm);
+    unlink(xen->corefile);
+    return -1;
+}
+
+/* ------------------------------------------------------------------ */
+
 static void collect_zombies(struct xenvm *xen)
 {
     int pid, status;
@@ -905,14 +975,14 @@ static void collect_zombies(struct xenvm *xen)
 
     for (;;) {
 	pid = waitpid(-1, &status, WNOHANG);
-	if (-1 == pid)
+	if (pid < 1)
 	    return;
 
 	if (pid == xen->qemu_pid) {
 	    xen->qemu_pid = 0;
 	    name = "qemu-dm";
 	}
-	
+
 	if (WIFEXITED(status)) {
 	    d1printf("child %d (%s) exited: %d\n",
 		     pid, name, WEXITSTATUS(status));
@@ -941,10 +1011,10 @@ int kvmbits_mainloop(struct xenvm *xen)
 	setup_regs(vcpu);
 	flush_regs(vcpu);
 	flush_sregs(vcpu);
-	pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);
     }
-    vcpu = xen->vcpu + 0;
+    kvmbits_vcpus_ready(xen);
 
+    vcpu = xen->vcpu + 0;
     for (;;) {
 	if (termsig)
 	    vm_kill(vcpu, "termsig", 0);
@@ -964,7 +1034,7 @@ int kvmbits_mainloop(struct xenvm *xen)
 	pthread_mutex_unlock(&xen->biglock);
 	rc = select(max+1, &rd, NULL, NULL, NULL);
 	pthread_mutex_lock(&xen->biglock);
-	
+
 	if (-1 == rc) {
 	    if (EINTR == errno) {
 		d1printf("%s: select: EINTR (got signal)\n", __FUNCTION__);
@@ -985,26 +1055,32 @@ int kvmbits_mainloop(struct xenvm *xen)
     return 0;
 }
 
-int kvmbits_vcpu_up(struct xenvcpu *vcpu)
-{
-    if (vcpu->thread)
-	return -EINVAL;
-    pthread_create(&vcpu->thread, NULL, vcpu_thread, vcpu);
-    return 0;
-}
-
-int kvmbits_vcpu_down(struct xenvcpu *vcpu)
-{
-    /* FIXME */
-    return -ENOSYS;
-}
-
 int kvmbits_vcpus_stop(struct xenvm *xen)
 {
+    xen->vcpus_stopped = 1;
+    d2printf("%s: stopped %d, ready %d\n", __FUNCTION__,
+             xen->vcpus_stopped, xen->ready_to_run);
+    if (!xen->ready_to_run)
+        return 0;
     return kvmbits_signal_vcpus(xen, SIGUSR1, 1);
 }
 
 int kvmbits_vcpus_cont(struct xenvm *xen)
 {
+    xen->vcpus_stopped = 0;
+    d2printf("%s: stopped %d, ready %d\n", __FUNCTION__,
+             xen->vcpus_stopped, xen->ready_to_run);
+    if (!xen->ready_to_run)
+        return 0;
+    return kvmbits_signal_vcpus(xen, SIGUSR2, 1);
+}
+
+int kvmbits_vcpus_ready(struct xenvm *xen)
+{
+    xen->ready_to_run = 1;
+    d2printf("%s: stopped %d, ready %d\n", __FUNCTION__,
+             xen->vcpus_stopped, xen->ready_to_run);
+    if (xen->vcpus_stopped)
+        return 0;
     return kvmbits_signal_vcpus(xen, SIGUSR2, 1);
 }
