@@ -87,13 +87,13 @@ static void net_tx_thread_gosleep(struct netdev *netdev)
 	    netdev->st.wakeups++;
     netdev->tx_work = 0;
 }
- 
+
 static void net_tx_response(struct netdev *netdev, netif_tx_request_t *txp, int8_t st)
 {
     RING_IDX i = netdev->tx_ring.rsp_prod_pvt;
     netif_tx_response_t *resp;
     int notify;
-    
+
     resp = RING_GET_RESPONSE(&netdev->tx_ring, i);
     resp->id     = txp->id;
     resp->status = st;
@@ -104,17 +104,17 @@ static void net_tx_response(struct netdev *netdev, netif_tx_request_t *txp, int8
     } else {
 	netdev->st.error_tx  += 1;
     }
-    
+
 #if 0
     if (txp->flags & NETTXF_extra_info)
 	RING_GET_RESPONSE(&netdev->tx_ring, ++i)->status = NETIF_RSP_NULL;
 #endif
-    
+
     netdev->tx_ring.rsp_prod_pvt = ++i;
     RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&netdev->tx_ring, notify);
     if (notify)
 	xc_evtchn_notify(netdev->xendev.evtchnd, netdev->xendev.local_port);
-    
+
     if (i == netdev->tx_ring.req_cons) {
 	int more_to_do;
 	RING_FINAL_CHECK_FOR_REQUESTS(&netdev->tx_ring, more_to_do);
@@ -153,7 +153,7 @@ static void *net_tx_thread(void *arg)
     RING_IDX rc, rp;
     void *page;
     int len, status;
-    
+
     d1printf("%s: start\n", __FUNCTION__);
     for (;;) {
 	if (netdev->xendev.state != XENDEV_CONNECTED)
@@ -161,8 +161,8 @@ static void *net_tx_thread(void *arg)
 
 	rc = netdev->tx_ring.req_cons;
 	rp = netdev->tx_ring.sring->req_prod;
-	rmb(); /* Ensure we see queued requests up to 'rp'. */
-	
+	xen_rmb(); /* Ensure we see queued requests up to 'rp'. */
+
 	while ((rc != rp)) {
 	    if (RING_REQUEST_CONS_OVERFLOW(&netdev->tx_ring, rc))
 		break;
@@ -239,7 +239,7 @@ static void net_rx_response(struct netdev *netdev,
     RING_IDX i = netdev->rx_ring.rsp_prod_pvt;
     netif_rx_response_t *resp;
     int notify;
-    
+
     resp = RING_GET_RESPONSE(&netdev->rx_ring, i);
     resp->offset     = offset;
     resp->flags      = flags;
@@ -252,12 +252,12 @@ static void net_rx_response(struct netdev *netdev,
 	netdev->st.packet_rx += 1;
 	netdev->st.bytes_rx  += size;
     }
-    
+
 #if 1
     d2printf("%s: idx %d, status %d, flags 0x%x\n",
 	     __FUNCTION__, i, resp->status, resp->flags);
 #endif
-    
+
     netdev->rx_ring.rsp_prod_pvt = ++i;
     RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&netdev->rx_ring, notify);
     if (notify)
@@ -273,7 +273,7 @@ static void *net_rx_thread(void *arg)
     RING_IDX rc, rp;
     int len, status;
     void *page;
-    
+
     d1printf("%s: start\n", __FUNCTION__);
     for (;;) {
 	if (netdev->xendev.state != XENDEV_CONNECTED)
@@ -283,9 +283,9 @@ static void *net_rx_thread(void *arg)
 
 	rc = netdev->rx_ring.req_cons;
 	rp = netdev->rx_ring.sring->req_prod;
-	rmb(); /* Ensure we see queued requests up to 'rp'. */
+	xen_rmb(); /* Ensure we see queued requests up to 'rp'. */
 	d2printf("%s: %d %d\n", __FUNCTION__, rc, rp);
-	
+
 	if (rc == rp || RING_REQUEST_CONS_OVERFLOW(&netdev->rx_ring, rc)) {
 	    char buf[PAGE_SIZE];
 	    d2printf("%s: no buffer, drop packet\n", __FUNCTION__);
@@ -349,18 +349,18 @@ static void netdev_nocsum(struct netdev *netdev, int nocsum)
     else
 	d1printf("ioctl TUNSETNOCSUM %d: ok\n", netdev->nocsum);
 }
-	    
+
 /* ------------------------------------------------------------- */
 
 static void net_setup_backend(struct netdev *netdev)
 {
     struct ifreq ifreq;
-    
+
     if (netdev->xendev.state != XENDEV_PROBED)
 	return;
     if (!netdev->xendev.online)
 	return;
-    
+
     /* read xenstore entries */
     if (NULL == netdev->mac)
 	netdev->mac = read_be_str(&netdev->xendev, "mac");
@@ -404,7 +404,7 @@ static void net_setup_backend(struct netdev *netdev)
 	    netdev_bridge_init(netdev);
 	if (-1 == ioctl(netdev->tun, TUNSETPERSIST, 0))
 	    d1printf("ioctl TUNSETPERSIST off: %s\n", strerror(errno));
-	    
+
 	write_be_str(&netdev->xendev, "hotplug-status", "connected");
     }
 
@@ -425,7 +425,7 @@ err_tun:
 static void net_connect_frontend(struct netdev *netdev)
 {
     int rx_copy;
-    
+
     if (netdev->xendev.state != XENDEV_INITIALISED)
 	return;
 
@@ -479,7 +479,7 @@ static void net_disconnect(struct netdev *netdev, enum xenbus_state state)
     void *dummy;
 
     if (netdev->xendev.state == XENDEV_DISCONNECTED)
-	return;
+        goto out;
     netdev->xendev.state = XENDEV_DISCONNECTED;
 
     if (netdev->tx_thread) {
@@ -509,6 +509,7 @@ static void net_disconnect(struct netdev *netdev, enum xenbus_state state)
 	netdev->rxs = NULL;
     }
 
+out:
     if (state)
 	change_state_xendev(&netdev->xendev, state);
 }
