@@ -37,7 +37,6 @@ static struct xenvm xen = {
     .qemu.upstream = 0,
     .qemu.path   = "/usr/" LIB "/xen/bin/qemu-dm",
     .qemu.kbd    = "en-us",
-    .qemu.serial = "none",
 };
 
 /* ------------------------------------------------------------------ */
@@ -55,7 +54,7 @@ static char search_exe[256];
 static void prepare_search_path(struct xenvm *xen, char *argv0)
 {
     char *h;
-    
+
     getcwd(search_cwd, sizeof(search_cwd));
     search_path[0] = search_cwd;
 
@@ -68,6 +67,8 @@ static void prepare_search_path(struct xenvm *xen, char *argv0)
 
 /* ------------------------------------------------------------------ */
 
+static int balloon_mb;
+
 #define MONITOR_GREETING  "xenner " VERSION " monitor\r\n"
 #define MONITOR_ERROR     "unknown command\r\n"
 #define MONITOR_ERROR_SUB "unknown sub-command\r\n"
@@ -103,7 +104,7 @@ static int qemu_monitor_info(struct xenvm *xen, int fd, char *line)
     static const char error[]  = MONITOR_ERROR_SUB;
     char msg[1024];
     int i, len;
-    
+
     if (0 == strcmp(line, "cpus")) {
 	for (i = 0, len = 0; i < xen->vcpus; i++) {
 	    len += snprintf(msg+len, sizeof(msg)-len,
@@ -114,6 +115,11 @@ static int qemu_monitor_info(struct xenvm *xen, int fd, char *line)
 	}
 	xwrite(fd, msg, strlen(msg));
 
+    } else if (0 == strcmp(line, "balloon")) {
+        /* FIXME: faked for libvirt */
+        snprintf(msg, sizeof(msg), "balloon: actual=%d\r\n", balloon_mb);
+	xwrite(fd, msg, strlen(msg));
+
     } else {
 	xwrite(fd, error, strlen(error));
 
@@ -143,6 +149,10 @@ static int qemu_monitor_line(struct xenvm *xen, int fd, char *line)
     } else if (0 == strcmp(line, "cont")) {
 	kvmbits_vcpus_cont(xen);
 
+    } else if (0 == strncmp(line, "balloon ", 8)) {
+        /* FIXME: faked for libvirt, should propagate to xenstore */
+        balloon_mb = atoi(line+8);
+
     } else if ((0 == strcmp(line, "system_shutdown"))   ||
                (0 == strcmp(line, "system_powerdown")))  {
 	kill(getpid(), SIGINT);
@@ -152,15 +162,20 @@ static int qemu_monitor_line(struct xenvm *xen, int fd, char *line)
 
     } else if (strlen(line)) {
 	xwrite(fd, error, strlen(error));
- 
+
     }
     xwrite(fd, prompt, strlen(prompt));
     return 0;
 }
 
-static int qemu_monitor_data(struct xenvm *xen, int fd, void *data)
+static int qemu_monitor_hello(struct xenvm *xen, int fd)
 {
     static const char msg[] = MONITOR_GREETING MONITOR_PROMPT;
+    return xwrite(fd, msg, strlen(msg));
+}
+
+static int qemu_monitor_data(struct xenvm *xen, int fd, void *data)
+{
     char buf[256];
     int rc, pos, i;
 
@@ -171,13 +186,12 @@ static int qemu_monitor_data(struct xenvm *xen, int fd, void *data)
 	    if (xen->mon_connected) {
 		d1printf("%s: disconnected\n", __FUNCTION__);
 		xen->mon_connected = 0;
+                qemu_monitor_hello(xen, fd);
 	    }
 	    return 0;
 	} else if (EAGAIN == errno) {
-	    if (!xen->mon_connected) {
+	    if (!xen->mon_connected)
 		d1printf("%s: connected\n", __FUNCTION__);
-		xwrite(fd, msg, strlen(msg));
-	    }
 	    xen->mon_connected = 1;
 	    return 0;
 	} else {
@@ -188,6 +202,9 @@ static int qemu_monitor_data(struct xenvm *xen, int fd, void *data)
 	d1printf("%s: read: EOF\n", __FUNCTION__);
 	break;
     default:
+        if (!xen->mon_connected)
+            d1printf("%s: connected\n", __FUNCTION__);
+        xen->mon_connected = 1;
 	buf[rc] = 0;
 	xwrite(fd, buf, rc); /* echo */
 	pos  = strlen(xen->mon_buffer);
@@ -239,8 +256,7 @@ static int qemu_monitor_config(struct xenvm *xen, char *arg)
     }
     if (dev) {
 	struct termios tio;
-	
-	d1printf("%s: %s\n", __FUNCTION__, dev);
+
 	tcgetattr (tty, &tio);
 	tio.c_lflag &= ~(ECHO|ICANON|ISIG);
 	tio.c_cc[VMIN] = 1;
@@ -256,7 +272,7 @@ static int qemu_monitor_setup(struct xenvm *xen, int tty)
 {
     xenner_register_cb(xen, "monitor",
 		       tty, NULL, qemu_monitor_data);
-    qemu_monitor_data(xen, tty, NULL);
+    qemu_monitor_hello(xen, tty);
     return 0;
 }
 
@@ -363,7 +379,7 @@ static int qemu_net_config_netbackd(struct xenvm *xen)
 	if (!arg)
 	    break;
 
-	d1printf("%s: %s\n", __FUNCTION__, arg);
+	d2printf("%s: %s\n", __FUNCTION__, arg);
 	if (0 == strncmp(arg, "none", 4)) {
 	    return 0;
 	}
@@ -371,7 +387,7 @@ static int qemu_net_config_netbackd(struct xenvm *xen)
 	    d1printf("%s: user network not implemented\n", __FUNCTION__);
 	    return 0;
 	}
-	
+
 	vlan = qemu_parse_int(arg, "vlan=", 0);
 	if (vlan >= sizeof(xen->nics)/sizeof(xen->nics[0])) {
 	    d1printf("%s: vlan number too big (%u, %zd max)\n", __FUNCTION__,
@@ -381,7 +397,7 @@ static int qemu_net_config_netbackd(struct xenvm *xen)
 	if (!xen->nics[vlan].configured)
 	    xen->nics[vlan].fd = -1;
 	xen->nics[vlan].configured++;
-	
+
 	if (0 == strncmp(arg, "tap", 3)) {
 	    xen->nics[vlan].fd = qemu_parse_int(arg, "fd=", -1);
 	    qemu_parse_string(arg, "ifname=", xen->nics[vlan].ifname,
@@ -392,16 +408,16 @@ static int qemu_net_config_netbackd(struct xenvm *xen)
 		    tap_fd_to_name(xen->nics[vlan].fd, xen->nics[vlan].ifname,
 				   sizeof(xen->nics[vlan].ifname));
 	    }
-	    d1printf("%s: %d: if=\"%s\"\n", __FUNCTION__, vlan,
+	    d2printf("%s: %d: if=\"%s\"\n", __FUNCTION__, vlan,
 		     xen->nics[vlan].ifname);
 	    /* ignored ... */
 	    qemu_parse_string(arg, "script=", dummy, sizeof(dummy));
 	}
-	
+
 	if (0 == strncmp(arg, "nic", 3)) {
 	    qemu_parse_string(arg, "macaddr=", xen->nics[vlan].mac,
 			      sizeof(xen->nics[vlan].mac));
-	    d1printf("%s: %d: mac=\"%s\"\n", __FUNCTION__, vlan,
+	    d2printf("%s: %d: mac=\"%s\"\n", __FUNCTION__, vlan,
 		     xen->nics[vlan].mac);
 	    /* ignored ... */
 	    qemu_parse_string(arg, "model=", dummy, sizeof(dummy));
@@ -441,14 +457,14 @@ static int qemu_disk_config_blkbackd(struct xenvm *xen)
     int i,n;
 
     n = sizeof(xen->qemu.disk)/sizeof(xen->qemu.disk[0]);
-    
+
     for (i = 0; i < n; i++) {
 	arg = xen->qemu.disk[i];
 	if (!arg)
 	    break;
 
-	d1printf("%s: %s\n", __FUNCTION__, arg);
-	
+	d2printf("%s: %s\n", __FUNCTION__, arg);
+
 	if (-1 != qemu_parse_string(arg, "if=", interface, sizeof(interface))) {
 	    if (0 != strcmp(interface, "xen")) {
 		d1printf("%s: if != xen, ignoring disk\n", __FUNCTION__);
@@ -463,20 +479,20 @@ static int qemu_disk_config_blkbackd(struct xenvm *xen)
 	    return 0;
 	}
 	xen->disks[index].configured++;
-	
+
 	qemu_parse_string(arg, "file=", xen->disks[index].diskimage,
 			  sizeof(xen->disks[index].diskimage));
 	if (-1 != qemu_parse_string(arg, "media=", buf, sizeof(buf)))
 	    if (0 == strcmp(buf, "cdrom"))
 		xen->disks[index].cdrom = 1;
-		
+
 	/* ignored ... */
 	qemu_parse_string(arg, "bus=", dummy, sizeof(dummy));
 	qemu_parse_string(arg, "cyls=", dummy, sizeof(dummy));
 	qemu_parse_string(arg, "snapshot=", dummy, sizeof(dummy));
 	qemu_parse_string(arg, "cache=", dummy, sizeof(dummy));
     }
-	
+
     return 0;
 }
 
@@ -484,21 +500,21 @@ static int qemu_disk_config_blkbackd(struct xenvm *xen)
 
 static int qemu_vnc_config(struct xenvm *xen, char *arg)
 {
-    d1printf("%s: %s\n", __FUNCTION__, arg);
+    d2printf("%s: %s\n", __FUNCTION__, arg);
     xen->qemu.vnc = arg;
     return 0;
 }
 
 static int qemu_kbd_config(struct xenvm *xen, char *arg)
 {
-    d1printf("%s: %s\n", __FUNCTION__, arg);
+    d2printf("%s: %s\n", __FUNCTION__, arg);
     xen->qemu.kbd = arg;
     return 0;
 }
 
 static int qemu_serial_config(struct xenvm *xen, char *arg)
 {
-    d1printf("%s: %s\n", __FUNCTION__, arg);
+    d2printf("%s: %s\n", __FUNCTION__, arg);
     xen->qemu.serial = arg;
     return 0;
 }
@@ -514,22 +530,28 @@ static void prepare_environment(struct xenvm *xen)
 	if (0 != access(buf, R_OK))
 	    continue;
 	setenv("LD_LIBRARY_PATH", search_path[i], 1);
-	d1printf("%s: LD_LIBRARY_PATH=%s\n", __FUNCTION__, search_path[i]);
+	d2printf("%s: LD_LIBRARY_PATH=%s\n", __FUNCTION__, search_path[i]);
 	break;
     }
 }
 
-static int do_fork_exec(struct xenvm *xen, char **argv)
+static void log_cmdline(struct xenvm *xen, const char *prefix, char **argv)
 {
-    char msg[1024];
+    char msg[4096];
     int i, len;
-    pid_t pid;
 
-    len = snprintf(msg, sizeof(msg), "%s: \"%s\"", __FUNCTION__, argv[0]);
+    len = snprintf(msg, sizeof(msg), "%s: \"%s\"", prefix, argv[0]);
     for (i = 1; argv[i]; i++)
 	len += snprintf(msg+len, sizeof(msg)-len, " \"%s\"", argv[i]);
-    d1printf("%s\n", msg);
-    
+    d2printf("%s\n", msg);
+}
+
+static int do_fork_exec(struct xenvm *xen, char **argv)
+{
+    pid_t pid;
+
+    log_cmdline(xen, __FUNCTION__, argv);
+
     pid = fork();
     switch (pid) {
     case -1: /* Oops */
@@ -547,25 +569,37 @@ static int do_fork_exec(struct xenvm *xen, char **argv)
 static void qemu_dm_setup(struct xenvm *xen)
 {
     char domid[16];
-    char *argv[] = {
-	"/usr/" LIB "/xen/bin/qemu-dm",
-	"-M", "xenpv",
-	"-k", xen->qemu.kbd,
-	"-vnc", xen->qemu.vnc,
-#if 0
-	"-serial", xen->qemu.serial,
-#endif
-	"-d", domid,
-	"-domain-name", xen->name,
-	NULL
-    };
-
-#if 1
-    if (0 == strcmp(xen->qemu.serial, "pty"))
-	fprintf(stderr, "char device redirected to /dev/null\n");
-#endif
+    char *argv[64];
+    int argc;
 
     snprintf(domid, sizeof(domid), "%d", xen->domid);
+    argc = 0;
+
+    argv[argc++] = xen->qemu.path;
+    argv[argc++] = "-M";
+    argv[argc++] =    "xenpv";
+    argv[argc++] = "-d";
+    argv[argc++] =    domid;
+    argv[argc++] = "-domain-name";
+    argv[argc++] =    xen->name;
+    if (xen->qemu.vnc) {
+	argv[argc++] = "-vnc";
+	argv[argc++] =    xen->qemu.vnc;
+	argv[argc++] = "-k";
+	argv[argc++] =    xen->qemu.kbd;
+        if (!xen->qemu.serial)
+            xen->qemu.serial = "vc";
+    } else {
+	argv[argc++] = "-nographic";
+        if (!xen->qemu.serial)
+            xen->qemu.serial = "stdio";
+    }
+    argv[argc++] = "-serial";
+    argv[argc++] =    xen->qemu.serial;
+    argv[argc++] = "-net";
+    argv[argc++] =    "none";
+
+    argv[argc++] = NULL;
     prepare_environment(xen);
     xen->qemu_pid = do_fork_exec(xen, argv);
 }
@@ -577,18 +611,13 @@ static void qemu_upstream_setup(struct xenvm *xen)
     int argc, i, n;
 
     snprintf(domid, sizeof(domid), "%d", xen->domid);
-    prepare_environment(xen);
 
     argc = 0;
     argv[argc++] = xen->qemu.path;
-#if 0
-    argv[argc++] = "-no-kvm";
-    argv[argc++] = "-kernel";
-    argv[argc++] =    "dummy";
-#endif
     argv[argc++] = "-M";
     argv[argc++] =    "xenpv";
-    argv[argc++] = "-domid";
+    argv[argc++] = "-xen-attach";
+    argv[argc++] = "-xen-domid";
     argv[argc++] =    domid;
     argv[argc++] = "-name";
     argv[argc++] =    xen->name;
@@ -597,6 +626,12 @@ static void qemu_upstream_setup(struct xenvm *xen)
 	argv[argc++] =    xen->qemu.vnc;
 	argv[argc++] = "-k";
 	argv[argc++] =    xen->qemu.kbd;
+        if (!xen->qemu.serial)
+            xen->qemu.serial = "vc";
+    } else {
+	argv[argc++] = "-nographic";
+        if (!xen->qemu.serial)
+            xen->qemu.serial = "stdio";
     }
     argv[argc++] = "-serial";
     argv[argc++] =    xen->qemu.serial;
@@ -608,7 +643,7 @@ static void qemu_upstream_setup(struct xenvm *xen)
 	argv[argc++] = "-net";
 	argv[argc++] = xen->qemu.nic[i];
     }
-    
+
     n = sizeof(xen->qemu.disk)/sizeof(xen->qemu.disk[0]);
     for (i = 0; i < n; i++) {
 	if (!xen->qemu.disk[i])
@@ -616,8 +651,9 @@ static void qemu_upstream_setup(struct xenvm *xen)
 	argv[argc++] = "-drive";
 	argv[argc++] = xen->qemu.disk[i];
     }
-    
+
     argv[argc++] = NULL;
+    prepare_environment(xen);
     xen->qemu_pid = do_fork_exec(xen, argv);
 }
 
@@ -715,7 +751,7 @@ static int run_pygrub(struct xenvm *xen, char *bootloader)
     if (NULL != (h = strstr(buf, "(args")))
 	if (1 != sscanf(h, "(args \"%512[^\"]", append))
 	    sscanf(h, "(args %512[^)]", append);
-    d1printf("%s: kernel=\"%s\" initrd=\"%s\" append=\"%s\"\n",
+    d2printf("%s: kernel=\"%s\" initrd=\"%s\" append=\"%s\"\n",
 	     __FUNCTION__, kernel, initrd, append);
 
     if (strlen(kernel)) {
@@ -732,20 +768,15 @@ static int run_pygrub(struct xenvm *xen, char *bootloader)
     return 0;
 }
 
-static void run_xenconsole(struct xenvm *xen)
-{
-    char domid[16];
-    char *argv[] = {
-	"/usr/" LIB "/xen/bin/xenconsole", domid,
-	NULL
-    };
-
-    snprintf(domid, sizeof(domid), "%d", xen->domid);
-    do_fork_exec(xen, argv);
-}
-
 /* ------------------------------------------------------------------ */
 
+static char *pidfile = NULL;
+
+static void pidfile_cleanup(void)
+{
+    unlink(pidfile);
+}
+
 static void sigsegv(int signo, siginfo_t *info, void *context)
 {
     struct statfs fs;
@@ -797,7 +828,7 @@ static void libvirt_caps(FILE *fp)
 {
     static struct kvm_callbacks dummy;
     void *kvm;
- 
+
     /* test for kvm */
     kvm = kvm_init(&dummy, NULL);
     if (NULL == kvm)
@@ -871,10 +902,12 @@ static void usage(FILE *fp)
 	    "   -serial     <config>     qemu style serial line (aka console) setup\n"
 	    "   -vnc        <display>    vnc display (needs qemu-dm from xen-tools)\n"
 	    "   -k          <kbd>        vnc clients keyboard map\n"
+	    "   -uuid       <uuid>       set domain uuid\n"
+	    "   -pidfile    <file>       specify pidfile\n"
 	    "\n"
 	    "options (xenner only):\n"
+	    "   -xen-domid  <id>         set domain id\n"
 	    "   -domid      <id>         set domain id\n"
-	    "   -uuid       <uuid>       set domain uuid\n"
 	    "   -network    <mac>,<br>   xenner style network setup\n"
 	    "   -debug      <nr>         set debug level\n"
 	    "   -logfile    <file>       set logfile\n"
@@ -882,7 +915,6 @@ static void usage(FILE *fp)
 	    "   -cap-on     <name>       enable kvm features\n"
 	    "   -cap-off    <name>       disable kvm features\n"
 	    "   -nostderr                send messages to logfile only\n"
-	    "   -xenconsole              connect to guests xen console\n"
 	    "   -dumpcore                don't catch fatal signals (sigsegv, sigbus)\n"
 	    "   -libvirt-caps            print libvirt capability xml snippet\n"
 	    "\n"
@@ -911,13 +943,13 @@ int main(int argc, char *argv[])
     char *uuid = NULL;
     char *bootloader = NULL;
     int monitor_fd = -1;
-    int xenconsole = 0;
     int dumpcore   = 0;
+    int populate   = 0;
     int opt, rc;
 
     xen.domid = getpid();
     INIT_LIST_HEAD(&xen.callbacks);
-    
+
     for (opt = 1; opt < argc; opt++) {
 	if (opt < argc-1) {
 	    /* options with one argument */
@@ -1000,7 +1032,8 @@ int main(int argc, char *argv[])
 		opt++; continue;
 
 	    /* the ones below are xenner only */
-	    } else if (0 == strcmp(argv[opt], "-domid")) {
+	    } else if (0 == strcmp(argv[opt], "-xen-domid") ||
+                       0 == strcmp(argv[opt], "-domid")) {
 		xen.domid = atoi(argv[opt+1]);
 		opt++; continue;
 
@@ -1023,6 +1056,10 @@ int main(int argc, char *argv[])
 		bootloader = argv[opt+1];
 		opt++; continue;
 
+	    } else if (0 == strcmp(argv[opt], "-pidfile")) {
+                pidfile = argv[opt+1];
+		opt++; continue;
+
 	    } else if (0 == strcmp(argv[opt], "-logfile")) {
 		openlog(&xen, argv[opt+1]);
 		opt++; continue;
@@ -1048,17 +1085,13 @@ int main(int argc, char *argv[])
 	    continue;
 
 	} else if (0 == strcmp(argv[opt], "-S")) {
-	    xen.start_stopped = 1;
+	    xen.vcpus_stopped = 1;
 	    continue;
 
 	} else if (0 == strcmp(argv[opt], "-nostderr")) {
 	    xen.nostderr = 1;
 	    continue;
 
-	} else if (0 == strcmp(argv[opt], "-xenconsole")) {
-	    xenconsole = 1;
-	    continue;
-
 	} else if (0 == strcmp(argv[opt], "-dumpcore")) {
 	    dumpcore = 1;
 	    continue;
@@ -1071,6 +1104,10 @@ int main(int argc, char *argv[])
 	    xen.qemu.upstream = 0;
 	    continue;
 
+	} else if (0 == strcmp(argv[opt], "-populate")) {
+            populate = 1;
+	    continue;
+
 	} else if (0 == strcmp(argv[opt], "-libvirt-caps")) {
 	    libvirt_caps(stdout);
 	    exit(0);
@@ -1095,6 +1132,22 @@ int main(int argc, char *argv[])
 	exit(0);
     }
 
+    loginit(colors);
+    log_cmdline(&xen, "started as", argv);
+
+    if (pidfile) {
+        char mypid[16];
+        int pid;
+	if (-1 == (pid = open(pidfile, O_WRONLY | O_CREAT | O_EXCL, 0600))) {
+	    fprintf(stderr,"open %s: %s\n",pidfile,strerror(errno));
+	    exit(1);
+	}
+	atexit(pidfile_cleanup);
+	sprintf(mypid,"%d",getpid());
+	write(pid,mypid,strlen(mypid));
+	close(pid);
+    }
+
     if (!xen.qemu.upstream) {
 	qemu_disk_config_blkbackd(&xen);
 	qemu_net_config_netbackd(&xen);
@@ -1104,15 +1157,10 @@ int main(int argc, char *argv[])
     if (NULL == xen.kernel || bootloader)
 	run_pygrub(&xen, bootloader);
     if (NULL == xen.kernel) {
-	logprintf(&xen, "No kernel specified and pygrub failed.\n"); 
+	logprintf(&xen, "No kernel specified and pygrub failed.\n");
 	exit(1);
     }
 
-    loginit(colors);
-    logprintf(&xen, "started as: ");
-    for (opt = 0; opt < argc; opt++)
-	logprintf(&xen, "\"%s\"%s", argv[opt], (opt == argc-1) ? "\n" : " "); 
-
     if (NULL == uuid) {
 	uuid_t u;
 	strcpy(xen.vm, "/vm/");
@@ -1133,8 +1181,21 @@ int main(int argc, char *argv[])
 	sigaction(SIGBUS,  &act, &old);
     }
 
-    /* init kvm */
-    rc = kvmbits_init(&xen);
+    /* init kvm #1 */
+    rc = kvmbits_init1(&xen);
+    if (0 != rc)
+	exit(1);
+
+    /* start other processes as needed */
+    if (-1 != monitor_fd)
+	qemu_monitor_setup(&xen, monitor_fd);
+    if (xen.qemu.upstream)
+	qemu_upstream_setup(&xen);
+    else
+	qemu_dm_setup(&xen);
+
+    /* init kvm #2 */
+    rc = kvmbits_init2(&xen, populate);
     if (xen.unlink_kernel)
 	unlink(xen.kernel);
     if (xen.unlink_ramdisk)
@@ -1144,16 +1205,6 @@ int main(int argc, char *argv[])
 	exit(1);
     }
 
-    /* start other processes as needed */
-    if (-1 != monitor_fd)
-	qemu_monitor_setup(&xen, monitor_fd);
-    if (xen.qemu.upstream)
-	qemu_upstream_setup(&xen);
-    else if (xen.qemu.vnc)
-	qemu_dm_setup(&xen);
-    if (xenconsole)
-	run_xenconsole(&xen);
-
     /* go boot and run guest */
     banner_print(&xen, "boot");
     kvmbits_mainloop(&xen);
