@@ -1,5 +1,4 @@
-#include <xen/elfnote.h>
-#include <xen/elfstructs.h>
+#include "xen/elfstructs.h"
 #include <xen/grant_table.h>
 
 #include "list.h"
